# Standard imports
import logging
import platform
from dataclasses import asdict
from datetime import datetime, timezone
from pathlib import Path
from zipfile import ZipFile

# Third party imports
import click
import tomlkit
from pydantic.dataclasses import dataclass

# Cellulose imports
from cellulose.artifact.cellulose_artifact import CelluloseArtifact, Metadata
from cellulose.configs.config import AllConfig
from cellulose.version import CELLULOSE_SDK_VERSION

logger = logging.getLogger(__name__)


@dataclass
class CelluloseArtifactManager:
    cellulose_artifact = CelluloseArtifact(
        file_paths=[],
        metadata=Metadata(
            title="",
            created_at=datetime.utcnow(),
            updated_at=datetime.utcnow(),
        ),
    )
    metadata_doc = tomlkit.document()

    def initialize_artifact_metadata(self):
        """
        This internal method initializes the title and created_at / updated_at
        parameters in the Cellulose artifact metadata file.
        """
        self.metadata_doc.add(
            tomlkit.comment(
                "This is an autogenerated TOML Cellulose artifact file."
            )
        )
        self.metadata_doc.add(
            tomlkit.comment("Please do not modify it directly.")
        )
        self.metadata_doc.add(tomlkit.nl())
        self.metadata_doc["title"] = "unknown"
        self.metadata_doc.add(tomlkit.nl())

        # Timestamps in UTC.
        utcnow = str(datetime.now(timezone.utc).isoformat())
        self.metadata_doc.add("created_at", utcnow)  # type: ignore
        self.metadata_doc.add("updated_at", utcnow)  # type: ignore
        self.metadata_doc["created_at"].comment("in UTC")  # type: ignore
        self.metadata_doc["updated_at"].comment("in UTC")  # type: ignore
        self.metadata_doc.add(tomlkit.nl())

        # Cellulose versions + platform info.
        env_info = tomlkit.table()
        env_info.add("cellulose_sdk_version", CELLULOSE_SDK_VERSION)
        env_info.add("python_version", platform.python_version())
        env_info.add("system", platform.system())
        env_info.add("release", platform.release())
        self.metadata_doc.add("environment", env_info)

    def generate_cellulose_artifact_filename(self, name: str) -> Path:
        """
        Internal method for generating Cellulose artifact filename.
        Returns a pathlib.Path
        """
        output_filename = Path("{name}.cellulose.zip".format(name=name))
        self.metadata_doc["title"] = str(output_filename)
        return output_filename

    def init(self):
        """
        This method initializes / reinitializes a new Cellulose file context.
        """
        self.initialize_artifact_metadata()

    def load(self, cellulose_artifact_path: str):
        """
        This method loads a current CelluloseArtifact.
        NOTE: It will also reinitialize the Cellulose file context. This means
        you may lose context on unsaved / unexported Cellulose files.
        """
        logger.info("Reinitialize the Cellulose context...")
        self.init()

        # TODO: Implement proper loading of Cellulose artifact from a
        #               file path.
        logger.info(
            "Loading Cellulose artifact: {cellulose_artifact_path}...".format(
                cellulose_artifact_path=cellulose_artifact_path
            )
        )
        # self.cellulose_artifact = cellulose_artifact

    def append(self, file: Path):
        """
        This method appends a file to the list of file paths in Cellulose
        artifact.

        Params
        -----
        file: Path - The file path to be embedded within Cellulose artifact.
        """
        # TODO: Verify if .load() file path to Cellulose artifact is
        # valid
        self.cellulose_artifact.file_paths.append(file)
        # Add metadata associated with it too.

    def export(self, name: str, target_directory: str):
        """
        This method takes in a name for the Cellulose output artifact and
        target directory for where to place it.

        Params
        ------
        name: str - The name of the Cellulose artifact.
        For example, "my_benchmarks" to export a "my_benchmarks.cellulose.zip"
        target_directory: str - The target directory for where to place this
        generated ".cellulose" artifact.
        """
        cellulose_artifact_filename = (
            self.generate_cellulose_artifact_filename(name=name)
        )
        output_cellulose_artifact_path = Path(target_directory).joinpath(
            cellulose_artifact_filename
        )
        logger.info(
            "The following files will be zipped into {output_file}:".format(
                output_file=output_cellulose_artifact_path
            )
        )
        for file_name in self.cellulose_artifact.file_paths:
            logger.info(file_name)

        # TODO: Update timestamps in the metadata file here.

        click.secho(
            "Packing Cellulose output artifacts...", fg="yellow", bold=True
        )
        with ZipFile(output_cellulose_artifact_path, "w") as zip:
            # writing each output file one by one
            for file in self.cellulose_artifact.file_paths:
                zip.write(file, Path(file).name)

            # writing the metadata file too.
            output_metadata_filename = "/tmp/{name}.toml".format(name=name)
            with Path(output_metadata_filename).open("w") as fout:
                fout.write(tomlkit.dumps(self.metadata_doc))
            zip.write(
                output_metadata_filename,
                Path(output_metadata_filename).name,
            )

        click.secho(
            "Results generated and exported to {output_cellulose_artifact_path}!".format(  # noqa: E501
                output_cellulose_artifact_path=output_cellulose_artifact_path
            ),
            fg="green",
            bold=True,
        )

    def update_artifact_metadata_section(
        self, key: str, model_config: AllConfig
    ):
        section = tomlkit.table()
        section.add("config", asdict(model_config))
        self.metadata_doc.add(key, section)
