# Standard imports
import logging
from pathlib import Path

# Third party imports
import torch
from pydantic.dataclasses import dataclass

# Cellulose imports
from cellulose.base.export import Export
from cellulose.onnx.checker import ONNXChecker
from cellulose.onnx.loader import ONNXLoader
from cellulose.output.output import ExportOutput, ONNXOutput, TorchScriptOutput

logger = logging.getLogger(__name__)


@dataclass
class PytorchExport(Export):
    @staticmethod
    def generate_absolute_tmp_directory(output_file_name: str) -> Path:
        """
        This function takes in an output filename, appends a /tmp directory to
        it then return the full path.
        """
        return Path("/tmp").joinpath(output_file_name)

    def export(
        self,
        torch_model,
        input,
    ) -> ExportOutput:
        """
        This method exports the model in various formats and returns the list
        of outputs accordingly.

        Params
        ------
        torch_model (nn.Module): The PyTorch model.
        input: Input tensors to the PyTorch model.
        """

        export_output = ExportOutput()
        if self.model_config.export_config.enable_onnx_export:
            msg = "Exporting TorchScript..."
            logger.info(msg)
            export_output.torchscript = self.export_torchscript(
                torch_model=torch_model,
            )
        if self.model_config.export_config.enable_onnx_export:
            msg = "Exporting ONNX..."
            logger.info(msg)
            export_output.onnx = self.export_onnx(
                torch_model=torch_model,
                input=input,
            )

        return export_output

    def export_torchscript(
        self,
        torch_model,
    ) -> TorchScriptOutput:
        """
        This method exports the current PyTorch model to TorchScript.
        """
        if self.model_config.decorator_config is None:
            error_msg = (
                "Expected type(decorator_config) to be DecoratorConfig, "
                "but got None instead"
            )
            logger.error(error_msg)
            raise Exception(error_msg)

        # Generate the output TorchScript file in a /tmp directory first.
        tmp_output_torchscript_file = self.generate_absolute_tmp_directory(
            output_file_name=self.model_config.output_config_internal.torchscript_output_filename,  # noqa: E501
        )

        # Script and save the model
        scripted_module = torch.jit.script(torch_model)
        scripted_module.save(tmp_output_torchscript_file)

        # then append it to the artifact_manager list later.
        self.artifact_manager.append(file=tmp_output_torchscript_file)

        # Return the output
        return TorchScriptOutput(
            torchscript_file=tmp_output_torchscript_file,
        )

    def export_onnx(
        self,
        torch_model,
        input,
    ) -> ONNXOutput:
        """
        This method exports the current PyTorch model to ONNX.

        Params
        ------
        torch_model (nn.Module): The PyTorch model.
        input: Input tensors to the PyTorch model.
        """
        if self.model_config.decorator_config is None:
            error_msg = "Expected type(decorator_config) to be DecoratorConfig, but got None instead"  # noqa: E501
            logger.error(error_msg)
            raise Exception(error_msg)

        # Generate the output ONNX file in a /tmp directory first.
        tmp_output_onnx_file = self.generate_absolute_tmp_directory(
            output_file_name=self.model_config.output_config_internal.onnx_output_filename,  # noqa: E501
        )

        # Export the model
        torch.onnx.export(
            torch_model,
            input,  # model input or tuple
            str(tmp_output_onnx_file),
            export_params=True,
            opset_version=10,
            do_constant_folding=True,
            input_names=self.model_config.decorator_config.input_names,
            output_names=self.model_config.decorator_config.output_names,
            dynamic_axes={
                "input": {0: "batch_size"},  # variable length axes
                "output": {0: "batch_size"},
            },
        )
        # then append it to the artifact_manager list later.
        self.artifact_manager.append(file=tmp_output_onnx_file)

        if self.model_config.export_config.enable_onnx_checks:
            msg = (
                "enable_onnx_checks set to True, "
                "running checks on exported artifacts..."
            )
            logger.info(msg)
            loader = ONNXLoader()
            onnx_model = loader.load(tmp_output_onnx_file)

            checker = ONNXChecker()
            checker.check(onnx_model)

        return ONNXOutput(onnx_file=tmp_output_onnx_file)
